/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "dist.h"
#include "my_rand.h"
#include "gen_dist.h"
#include "pool.h"
#include "sb.h"
#include "MMM.h"
#include "Msbi.h"
#include "rand_gen.h"
#include "do_test.h"

type_data this_one;

coord	**P;

#define N_ELEMS(x) ((int)(sizeof(x)/sizeof((x)[0])))

int dims[] = {2, 4, 8, 16, 24, 32};

int dim_pool[30];
int num_sites[] = {256, 512, 1024, 2048, 4096, 8192, 16384, 32768};

double alphas[] = {0.1, 0.5, 1};

struct {char *nam; search_sb_f* func;} search_sbs[] = {
	{"d_order", &search_sb_d_order}
	,{"i_order", &search_sb_i_order}
	,{"stack", &search_sb_stack}
	,{"semi_h", &search_sb_semi_heap}
	,{"no_h", &search_sb_no_heap}
};


static void generate_data_r(
  coord **P, int i_dims, int num_tot, rand_pt_f rr, double rr_param) {
	int i,j, dim = dims[i_dims];

	for (j=0; j<num_tot; j++) {
		P[j] = (coord*)pool_new(dim_pool[i_dims]);
		for (i=0;i<dim;i++) rr(P[j], dim, rr_param);
	}
	this_one.P = P;
	this_one.dim = dim;
	this_one.num_P = num_tot;
}

static void gen_uniform_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_uniform_pt, 0);}

static void gen_laplace_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_laplace_pt, 0);}

static void gen_gauss_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_gauss_pt, 1);}

static void gen_co_laplace_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_co_laplace_pt, 0.8);}

static void gen_co_gauss_data(int i_dims, int num_tot)
	{generate_data_r(P, i_dims, num_tot, rand_co_gauss_pt, 0.8);}


int num_centers = 10;
static void gen_clus_gauss_data(int i_dims, int num_tot) {
	int i,j, dim = dims[i_dims];
	coord **centers = (coord**)calloc(num_centers+1, sizeof(coord*));

	generate_data_r(centers, i_dims, num_centers, rand_uniform_pt, 0);
	generate_data_r(P, i_dims, num_tot, rand_gauss_pt, 1.0/2/num_centers);
		/* this_one is now correctly set */

	for (j=0; j<num_tot; j++) {
		double rr = my_rand();
		int jp = (int)(rr*num_centers);
		for (i=0;i<dim;i++) P[j][i] += centers[jp][i]; 
	}
}

typedef void gen_data_f(int, int);



struct {char *nam; gen_data_f* func;} gen_data_fs[] = {
	 {"uniform_data", &gen_uniform_data}
	,{"laplace_data", &gen_laplace_data}
	,{"gauss_data", &gen_gauss_data}
	,{"co_gauss_data", &gen_co_gauss_data}
	,{"co_laplace_data", &gen_co_laplace_data}
	,{"clus_gauss_data", &gen_clus_gauss_data}
};

static void panic(char *fmt, ...) {
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	fflush(stderr);
	va_end(args);

	abort();
}

int count_lines(const char *f) {
	FILE *F = fopen(f, "r");
	int c=0;
	char buf[1000];

	if (!F) panic("whoops, can't open file for count");

	while(fgets(buf, sizeof(buf), F)) c++;

	return c;
}

main(int argc, char** argv) {

	size_t i_dims, i_num_sites, i_search_sbs, i_alphas, i_gen_data_fs;
	size_t trial_count, cc=0;
	char packing = 1;
	int nt, ntd=100, L=5;
	double end_frac=0.1, num_reps = 5;
	sbp sbt;
	size_t i, maxn, num_P;

	trial_count = count_lines("trials");

	printf("trialcount/num_reps=%d\n", trial_count/num_reps);

	switch (argc) {
		case 6: num_reps = (double)atoi(argv[5]);
		case 5: ntd = atoi(argv[4]);
		case 4:	end_frac= atof(argv[3]);	
		case 3: packing = atoi(argv[2]);	
		case 2: sbtesttype = (sbtesttypes)atoi(argv[1]); /* intentional fallthru */
	}

	init_my_rand();

	printf("sbtesttype=%s\n",
		(sbtesttype==test_closest) ? "test_closest" 
								   : ((sbtesttype==test_k_nn) ? "k_nn" : "fixed_radius"));

	maxn = ntd + num_sites[(N_ELEMS(num_sites)-1)];
	shufmat = (size_t*)malloc(maxn*sizeof(size_t));
	for (i=0; i<maxn; i++) shufmat[i] = i;

	P = (coord**)calloc(maxn+1, sizeof(coord*));

	ITERATE(dims) {
		dim_pool[i_dims] = pool_init(dims[i_dims]*sizeof(coord));
		ITERATE(gen_data_fs) {
			if (num_reps*cc>=trial_count)
				printf("dataset=euclid_%d_%s\n", dims[i_dims],
					gen_data_fs[i_gen_data_fs].nam);
			ITERATE(num_sites) {
				if (num_reps*cc++ < trial_count)
					continue;
				num_P = num_sites[i_num_sites];
				printf("num_sites=%d\n", num_P);
				nt = ntd;
				printf("num_tests=%d\n", nt);
				gen_data_fs[i_gen_data_fs].func(i_dims, num_P+nt);
				use_data(&this_one);
				activity = prep;
				memset(dcounts, 0, sizeof(dcounts));
				sbt = build_sbp(my_dist, num_P, (float)end_frac, packing, this_dist2);
				print_prep(sbt);
				if (sbtesttype==test_closest) ITERATE(search_sbs) {
					printf("search_method=%s\n", search_sbs[i_search_sbs].nam);
					ITERATE(alphas)  {
						printf("alpha=%0.1f\n", alphas[i_alphas]);
						fflush(stdout);
						activity = query;
						test_sbt(sbt, num_P, num_P, nt,
							(float)alphas[i_alphas],
							search_sbs[i_search_sbs].func, 0);
					}
				} else {
					fflush(stdout);
					activity = query;
					test_sbt(sbt, num_P, num_P, nt, 0, NULL, 0);
				}
			}
		}
		pool_purge(dim_pool[i_dims]);
	}
	pool_purge_all();
	return 0;
}
