/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/


#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "heap.h"

typedef struct SD {
    int point;
    float key;
} SD;


void print_heap_recur(heap *h, size_t j, size_t level) {
    size_t i;
    if (j>=h->num_entries) return;
    print_heap_recur(h, 2*j+1, level+1);
    for (i=0; i<level; i++) printf(" ");
    printf("%d : %G\n", h->h[j].point, h->h[j].key);
    print_heap_recur(h, 2*j+2, level+1);
}

void print_heap(heap *h) {

    size_t j;

    for (j=0; j<h->num_entries; j++) {
        printf("%d, %G  ", h->h[j].point, h->h[j].key);
    }
}

#define SD_swap(a, b) {SD t;  t = *(a); *(a) = *(b);  *(b) = t;}                    

void insert_SD(SD* hhj, size_t point, float key) {
    hhj->point = point;
    hhj->key = key;
}

static char check_heap(heap *h, size_t j) {

    if (j>=h->num_entries) return 1;
    if (2*j+1 < h->num_entries && h->h[j].key < h->h[2*j+1].key) {
        printf("heap conditions fail at %d\n", 2*j+1);
        return 0;
    }
    if (2*j+2 < h->num_entries && h->h[j].key < h->h[2*j+2].key) {
        printf("heap conditions fail at %d\n", 2*j+2);
        return 0;
    }
    if (!check_heap(h, 2*j+1)) return 0;
    if (!check_heap(h, 2*j+2)) return 0;
    return 1;
}



void sift_down(heap *h, size_t k) {

    SD    *hh = h->h, *c1, *c2=0, *max;
    size_t    kk;


    while (1) {
        kk = 2*k+1;
        if (kk >= h->num_entries) return; /* h->num_entries is even,
                            so we know that hh[k]
                            has no children */

        max = hh + k;
        c1 = hh + kk;
        if (c1->key > max->key) max = c1;
        c2 = c1+1;
        if (c2 < hh + h->num_entries && c2->key > max->key) max = c2;
        if (max == hh + k) return;
        SD_swap(max, hh+k)
        k = max - hh;
    }
}

void sift_up(heap *h, size_t k) {
    SD *hh = h->h;
    size_t j,jj;
    if (k>h->num_entries) return; /* why not >=? */
    for (j=k; j>0; j=jj) {
        if (h->h[jj=(j-1)/2].key >=hh[j].key) break;
        SD_swap(hh+j, hh+jj);
    }
}

static void realloc_heap(heap *h) {
    h->h =  (SD*)realloc(h->h, h->num_allocated*sizeof(*h->h));
	assert(h->h);
}

static void insert_list(heap *h, size_t a, float d) {
    size_t i;

    if (h->num_entries >= h->num_allocated) {
		if (h->num_allocated==0)
			h->num_allocated = 1;
		else
			h->num_allocated *= 2;
        realloc_heap(h);
        for (i=h->num_entries; i< h->num_allocated; i++)
            h->h[i].key = -FLT_MAX;
    }

    insert_SD(h->h + h->num_entries++, a, d);
}

void insert_heap(heap *h, size_t a, float d) {
    insert_list(h,a,d);
    sift_up(h, h->num_entries-1);
}

static void adjust_heap_size(heap *h) {
    if (h->num_entries <= h->num_allocated/4) {
        h->num_allocated /= 2;
        if (h->num_allocated==0)
			h->num_allocated = 1;
        realloc_heap(h);
    }
}


void delete_heap(heap *h, size_t j) {
    SD *hh = h->h;

    assert(j<h->num_entries);

    h->num_entries--;
    hh[j] = hh[h->num_entries];
    hh[h->num_entries].key = -FLT_MAX;

    adjust_heap_size(h);

    sift_up(h, j);
    sift_down(h,j);
}


void init_heap(heap *h) {memset(h, 0, sizeof(h[0]));}

void free_heap(heap* h) {
    free(h->h);
    free(h);
}


