/****************************************************************************
* Copyright (C)2002 Lucent Technologies                                     *
* (clarkson@research.bell-labs.com http://cm.bell-labs.com/who/clarkson)    *
*                                                                           *
* This program is free software; you can redistribute it and/or             *
* modify it under the terms of the GNU General Public License               *
* as published by the Free Software Foundation; either                      *
* version 2 of the License, or (at your option) any later version.          *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License for more details.                              *
*                                                                           *
* You should have received a copy of the GNU General Public License         *
* along with this program; if not, write to the Free Software               *
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,                 *
* or visit http://www.gnu.org/licenses/gpl.html                             *
*****************************************************************************/



/***********************************************************************
* sb structure
***********************************************************************/


#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "pool.h"
#include "heap.h"
#include "sb.h"
#include "Msbi.h"
#ifdef DO_STATS
#include "MMM.h"
#include "do_test.h"
#endif

float stat_change_nn;
float stat_max_heap_entries;
float stat_search_loop;
activities activity; /* for performance statistics */

static char show_progress; /* if !=0, show progress in building sb */

static int alist_str_poolnum = -1;

static heap msh; /* msh is heap of sites; site i has key==current max dist to i from sites with i closest */


/* the site j has already been considered for current search */
#define MARK_SEEN(sbt, j) {sbt->seen_for[j] = 1; sbt->seens[sbt->ns++] = j;} 

void reset_seens(sbp sbt)
{
    if (sbt->ns < sbt->n/50) {/* from comparative timings */
		while (sbt->ns>0) sbt->seen_for[sbt->seens[--sbt->ns]] = 0;
    } else memset(sbt->seen_for, 0, sbt->n*sizeof(char));
    sbt->ns = 0;
}

static void reset_aux(sbp sbt)
{
	reset_seens(sbt);
	if (sbt->psh->h)
		free(sbt->psh->h);
	init_heap(sbt->psh);
}


sb *init_sb(sb_distance dist, size_t n, float end_frac,
			char sb_flag, sb_distance dist2)
{
    sb *sbt = (sb*)malloc(sizeof(sb));
    make_storage(sbt->offset, size_t, n+1);
    sbt->dist = dist;
	sbt->dist2 = dist2 ? dist2 : dist;
    sbt->n = n;
    sbt->end_thresh = (size_t)(end_frac*n + 0.5);
	if (sbt->end_thresh==0)
		sbt->end_thresh = 1;
	
    if (!sb_flag) return sbt;
	
    make_storage(sbt->seen_for, char, n);
    make_storage(sbt->seens, size_t, n);
    make_storage(sbt->cur_touching, size_t, n);
    make_storage(sbt->dist2q, float, n);
    make_storage(sbt->ins_num, size_t, n);
    sbt->ns = 0;
	
    return sbt;
}

typedef struct site_data_sb_str {
    heap    l1; /* heap of sites with this one closest, farthest at the top */
    int    last_l1_visit, last_pending;
    alist    last_affL, affL; /* last_affL is the last entry of the affL list */
    SD    nn;
} site_data_sb_str;

site_data_sb_str *sd_sb;

#define farthest_owned(i) sd_sb[i].l1.h[0].key



void show_sb(sb *sbt) {
	
    size_t i, j;
	
    for (i=0; i< sbt->n; i++) {
        if (sbt->offset[i]==sbt->offset[i+1]) continue;
        printf("%d: ", i); 
        for (j=sbt->offset[i]; j<sbt->offset[i+1]; j++)
            printf("%d ", sbt->DL[j]);
        printf("\n");
    }
    fflush(stdout);
}


ddats edists;

static void update_ddat(ddats *ed, float old, float ne)
{
    if (ed->min.old>old) ed->min.old = old;
    if (ed->max.old<old) ed->max.old = old;
    if (ed->min.nnew>ne) ed->min.nnew = ne;
    if (ed->max.nnew<ne) ed->max.nnew = ne;
}


static void init_ddat(ddats *ed)
{
    ed->max.old = ed->max.nnew = 0;
    ed->min.old = ed->min.nnew = FLT_MAX;
}

SD no_point = {-1, -FLT_MAX};

/* find sites that had m nearest, and now have i nearest */
/* need consider only sites farther than X from m  */
/* h is index into heap of sites with m nearest */
/* return value is nonzero iff some site now has i nearest */

static char search4nns(sb* sbt, size_t i, size_t m, size_t h, float X)
/* checking if sites with m nearest before have i nearest now             */
{
    heap *ah = &sd_sb[m].l1; /* heap of sites, farthest from m on top */
    SD cand, *ahh = ah->h;
    char rval = 0;
    size_t al;
    float dci;
    int *v;
	
    while (h < ah->num_entries && ahh[h].key > X)  {
		int visited;
		cand = ahh[h];
        v = &sd_sb[cand.point].last_l1_visit;
		visited = (*v==(int)i) ? 1 : 0; /* have we already looked at cand.point for i? */
		*v = i; 
		
        if (visited || (dci=sbt->dist(cand.point, i)) >= cand.key) {/* can ignore this one */
            sift_up(ah, h); /* repair any deletions */
            rval |= search4nns(sbt, i, m, 2*h+1, X);
            rval |= search4nns(sbt, i, m, 2*h+2, X);
            return rval;
        }
		/*  cand.point has i as nn */
		update_ddat(&edists, cand.key, dci);
		
        sd_sb[cand.point].nn.point = i;
        sd_sb[cand.point].nn.key = dci;
		rval = 1;
		
        insert_heap(&sd_sb[i].l1, cand.point, dci); /* insert cand into i's heap */
		
        al = --(ah->num_entries); /* delete cand from m's heap */
        ahh[h] = ah->h[al];
		ahh[al] = no_point;
        sift_down(ah, h);   /* now ahh[h] may have larger key than par */
							/* we don't sift up, since we need to check this
								entry */
    };
    return rval;
}

static alist new_alist(sb *sbt, size_t m, site_data_sb_str *sa)
{
	alist v = (alist)pool_new(alist_str_poolnum);
    sbt->offset[m+1]++;
    if (sa->last_affL)
        sa->last_affL->next = v;
    else {
        sa->affL = v;
    }
    sa->last_affL = v;
	return v;
}


/* find sites with im nearest; look on touching list for m.
A site j on m's touching list took sites away from m when j was added;
recursively look at all such j
*/
static void inverse_search(sb* sbt, size_t im, size_t m, heap *pmsh) {
    float d, fnd; 
    alist ma,v;
    size_t j;
    site_data_sb_str *sa;
	
    assert(im!=m);
    d = sbt->dist(m, im);
    for (ma=sd_sb[m].affL; ma; ma = ma->next) {
        if (2*ma->fnn <= d) 
			break;
        if (sd_sb[j=ma->site].last_pending == (int)im)
			continue; /* we already considered j */
        if (2*ma->nn.max.old < d)
			continue; /* max dist from ma->site to sites it took from m is small */
        sd_sb[j].last_pending = im;
        inverse_search(sbt, im, j, pmsh);
    }
	
    if (2*farthest_owned(m) <= d)
		return; /* there can't be a site with m formerly closest, and im closer */
	
    sa = sd_sb + m;
    fnd = farthest_owned(m);
    assert(im!=m);
	init_ddat(&edists);
    if (!search4nns(sbt, im, m, 0, d/2))
		return;
	/* add im to is touched list of m */
    if (pmsh!=NULL)
        insert_heap(pmsh, m, farthest_owned(m));
	
	v = new_alist(sbt, m, sa);
    v->site = im;
    v->nn = edists;
    v->fnn = fnd;
}



sb *build_sb(sb_distance dist, size_t num_sites) {
	return build_sbp(dist, num_sites, (float)0.1, 1, 0);
}

size_t inserted(sb *sbt, size_t i) {return sbt->ins[i];}

size_t size_lists(sb *sbt) {return sbt->offset[sbt->n];}


static void init_sd_sb(sb *sbt, int num_sites)
{
	int i;
    make_storage(sd_sb, site_data_sb_str, num_sites);
	
    for (i=1; i<num_sites; i++) {
        float d0 = sbt->dist(0,i);
        sd_sb[i].last_pending = sd_sb[i].last_l1_visit= -1;
        insert_heap(&sd_sb->l1, i, d0);
        sd_sb[i].nn.point = 0;
        sd_sb[i].nn.key = d0;
    }
}


static void insert_main(sb *sbt, heap *pmsh, size_t i)
{
	size_t im;
	
	if (show_progress && i%100==0) {printf(".");fflush(stdout);}
	if (pmsh==NULL) {
		im = sbt->ins[i] = sbt->ins_num[i] = i; 
		inverse_search(sbt, im, 0, NULL);
		if (0) assert(sd_sb[im].nn.point==(int)im);
	} else {
		while (topp(pmsh).key > farthest_owned(topp(pmsh).point))
			delete_heap(pmsh, 0);
		sbt->ins[i] = im = top(sd_sb[topp(pmsh).point].l1).point; 
		sbt->ins_num[im] = i; 
		inverse_search(sbt, im, 0, pmsh);
		insert_heap(pmsh, im, farthest_owned(im));
	}
}

static void insert_rest(sb *sbt, size_t j)
{
	size_t i = sbt->ins[j];
    site_data_sb_str *si = &sd_sb[i];
    heap *hi = &si->l1;
	alist v;
	
	while (hi->num_entries!=0) {
		if ((int)i==topp(hi).point || topp(hi).key==0) {
			delete_heap(hi,0);
			continue;
		}
		v = new_alist(sbt, i, si);
		v->site = topp(hi).point;
		v->nn.max.old = v->nn.min.old = v->fnn = topp(hi).key;
		v->nn.min.nnew = v->nn.min.nnew = 0;
		sbt->ins_num[v->site] = sbt->end_thresh;
		delete_heap(hi,0);
	}
}


/* translate from a list format to a more compact one */
static void sb_from_sd_sb(sb *sbt, size_t i)
{
    size_t j;
    alist ia = sd_sb[i].affL;
	
    for (j=sbt->offset[i]; j<sbt->offset[i+1]; j++) {
        assert(ia);
        sbt->DL[j] = ia->site;
        sbt->dDL[j] = ia->fnn;
        sbt->nn[j] = ia->nn;
        ia = ia->next;
    }
	
    assert(!ia);
}


sb *build_sbp(sb_distance dist, size_t num_sites,
			  float end_frac, char packing, sb_distance dist2)
{
    sb *sbt;
    size_t i, *so;
	
	activity = prep;
    assert(dist!=0);
    init_heap(&msh);
    if (alist_str_poolnum==-1)
        alist_str_poolnum = pool_init(sizeof(alist_str));
	
    sbt = init_sb(dist, num_sites, end_frac, 1, dist2);
    make_storage(sbt->ins, size_t, num_sites+1);
	init_sd_sb(sbt, num_sites);
	
 	make_storage(sbt->psh, heap, 1);
	init_heap(sbt->psh);
	
	insert_heap(&msh, 0, farthest_owned(0) );
	
    for (i=1; i<sbt->end_thresh; i++)
		insert_main(sbt, (packing ? &msh : NULL), i);
	
    for (i=0; i<sbt->end_thresh; i++)
		insert_rest(sbt, i);
	
    for (so = sbt->offset,i=1; i<=sbt->n; i++,so++)
		*(so+1) += *so;
	
    make_storage(sbt->DL, size_t, *so);
    make_storage(sbt->dDL, float, *so);
	make_storage(sbt->nn, ddats, *so);
	
	for (i=0; i<sbt->n; i++)
		sb_from_sd_sb(sbt, i);
	
    for (i=0; i<num_sites; i++)
		free(sd_sb[i].l1.h);
	
    free(msh.h);
    pool_purge(alist_str_poolnum);
    free(sd_sb);
	
    return sbt;
}

void free_sb(sbp sbt) {free_sb_i(sbt, 1);}

void free_sb_i(sbp sbt, int sb_flag) {
    if (!sbt) return;
    if (sbt->offset) free(sbt->offset);
    if (!sb_flag) {free(sbt); return;}
    if (sbt->seen_for) free(sbt->seen_for);
    if (sbt->seens) free(sbt->seens);
    if (sbt->cur_touching) free(sbt->cur_touching);
    if (sbt->dist2q) free(sbt->dist2q);
    if (sbt->ins_num) free(sbt->ins_num);
    if (sbt->nn) free(sbt->nn);
    if (sbt->DL) free(sbt->DL);
    free(sbt);
}







/***********************************************************************
*                 search procedures for sb data structure              *
***********************************************************************/

size_t search_sb(sb* sbt, size_t q, float alpha)
{
	if (sbt==0 || q>sbt->n || alpha > 1 || alpha <0) {
		fprintf(stderr, "search_sb: failed search: null sb*");
		exit(1);
	}
	return search_sb_d_order(sbt, q, alpha);
}

/* search_sb_stack */

size_t search_sb_stack(sb* sbt, size_t q, float alpha)
{
	size_t j, k, m;
    size_t *ps, ps_top=1, *have_d;
    int nn_p;
    SD q_nn;
		
    make_storage(ps, size_t, sbt->n+1);
    make_storage(have_d, size_t, sbt->n+1);
    
    q_nn.key = sbt->dist2q[0] = sbt->dist(0,q);
    q_nn.point=0;
	
    while (ps_top>0) {
        m = ps[--ps_top];
        nn_p = -1;
        for (k=sbt->offset[m]; k<sbt->offset[m+1]; k++) {
            stat_inc_search_loop;
            if (sbt->dDL[k] + alpha*q_nn.key < sbt->dist2q[m]) break;
            if (sbt->seen_for[j=sbt->DL[k]])
                continue;
            if (0 && q_nn.point < (int)m && sbt->nn[k].min.old > (1+alpha)*q_nn.key)
                continue;
            if (0 && q_nn.point < (int)j && sbt->nn[k].min.nnew > (1+alpha)*q_nn.key)
                continue;
            if (sbt->nn[k].max.old + alpha*q_nn.key < sbt->dist2q[m])
                continue;
			
            if (!have_d[j]) {
                sbt->dist2q[j] = sbt->dist(j, q);
                have_d[j] = 1;
            }
            if (sbt->dist2q[j]<q_nn.key) {
                stat_inc_change_nn;
                q_nn.key = sbt->dist2q[j];
                q_nn.point = j;
                nn_p = ps_top-1;
            } else if (sbt->nn[k].max.nnew + alpha*q_nn.key < sbt->dist2q[j])
                continue;
            MARK_SEEN(sbt, j);
            ps[ps_top++] = j;
        }
        if (nn_p!=-1)
		{size_t s=ps[ps_top-1]; ps[ps_top-1] = ps[nn_p]; ps[nn_p] = s;}
		
    }
    reset_seens(sbt);
    free(ps); free(have_d);
    return q_nn.point;
}



static void search_sb_no_heap_inner(sb* sbt, size_t q, size_t m, float alpha, SD *q_nnp) {
	
    size_t ma, j;
    float d;
	
    for (ma=sbt->offset[m]; ma<sbt->offset[m+1]; ma++) {
        stat_inc_search_loop;
        if (sbt->dDL[ma] + alpha*q_nnp->key < sbt->dist2q[m]) break;
        j = sbt->DL[ma];
        if (sbt->seen_for[j]) continue;
        if (sbt->nn[ma].max.old + alpha*q_nnp->key < sbt->dist2q[m])
            continue;
        sbt->dist2q[j] = d = sbt->dist(q,j);
        if (d < q_nnp->key) {q_nnp->key = d; q_nnp->point = j;}
        MARK_SEEN(sbt, j)
			search_sb_no_heap_inner(sbt, q, j, alpha, q_nnp);
    }
}

/* search_sb_no_heap */

/* static heap sh; */


size_t search_sb_no_heap(sb* sbt, size_t q, float alpha)
{
	SD q_nn;
	
    q_nn.point = 0;
    q_nn.key = sbt->dist2q[0] = sbt->dist(0,q);
    search_sb_no_heap_inner(sbt, q, 0, alpha, &q_nn);    
	
    reset_aux(sbt);
    return q_nn.point;
}




/* search_sb_semi_heap */

size_t search_sb_semi_heap(sb* sbt, size_t q, float alpha)
{
	
    size_t j, k, m;
    SD q_nn;
    float d;
		
    q_nn.key = sbt->dist2q[0] = sbt->dist(0,q);
    q_nn.point=0;
    insert_heap(sbt->psh, 0, -sbt->dist2q[0]);
	
    while (sbt->psh->num_entries >0) {
        m = topp(sbt->psh).point;
        delete_heap(sbt->psh, 0);
        for (k=sbt->offset[m]; k<sbt->offset[m+1]; k++) {
            stat_inc_search_loop;
            if (sbt->dDL[k] + alpha*q_nn.key < sbt->dist2q[m]) break;
			/*            if (q_nn.point < m && 3*q_nn.key < sbt->dist2q[m]) break;*/
            if (sbt->seen_for[j=sbt->DL[k]]) continue;
            stat_inc_search_loop;
            MARK_SEEN(sbt, j);
			d = sbt->dist2q[j] = sbt->dist(j, q);
            insert_heap(sbt->psh, j, -d);
            stat_update_max_heap_entries((float)sbt->psh->num_entries);
            if (d<q_nn.key) {
                stat_inc_change_nn;
                q_nn.key = d;
                q_nn.point = j;
            }
        }
    }
	
    reset_aux(sbt);
    return q_nn.point;
}






/* search sb, next touching site considered has smallest insertion number */

size_t search_sb_i_order(sb* sbt, size_t q, float alpha)
{
	size_t j, m, ma;
    float d = 0.0;
	SD q_nn = {0, FLT_MAX};
    j = ma = 0;

    while (1) {
        if (  !sbt->seen_for[j]
			&& sbt->nn[ma].max.old + alpha*q_nn.key > d 
			&& ((1+2*alpha)*q_nn.key > d || sbt->ins_num[j] <= sbt->ins_num[q_nn.point])
			) {
            sbt->dist2q[j] = d = sbt->dist(q,j);
            if (d < q_nn.key) {q_nn.key = d; q_nn.point = j;}
            ma = sbt->cur_touching[j] = sbt->offset[j];
            if (ma<sbt->offset[j+1]) {
                insert_heap(sbt->psh, j, -(float)sbt->ins_num[sbt->DL[ma]]);
                MARK_SEEN(sbt, j);
            }
        }
        if (sbt->psh->num_entries==0) break;
        do {
            stat_inc_search_loop;
            m = topp(sbt->psh).point;
            d = sbt->dist2q[m];
            ma = (sbt->cur_touching[m])++;
            delete_heap(sbt->psh, 0);
            if (ma>=sbt->offset[m+1]
				|| sbt->dDL[ma] + alpha*q_nn.key < d) continue;
            j = sbt->DL[ma];
            if (ma<sbt->offset[m+1]-1)
                insert_heap(sbt->psh, m, -(float)sbt->ins_num[sbt->DL[ma+1]]);
            stat_update_max_heap_entries(sbt->psh->num_entries);
            break;
        } while (sbt->psh->num_entries>0);
    }
	
    reset_aux(sbt);

    return q_nn.point;
}



/*
keep a heap of pending, based on dist to query;
always get touching site for pending site nearest to query
similar to search_sb_i_order, but: need additional condition
for (1+2*alpha)*q_nn.key > d test, don't need dist2q, and don't
need to delete and then re-insert, since key doesn't change
*/

size_t search_sb_d_order(sb* sbt, size_t q, float alpha)
{
    size_t pj=0, touch_index=0;
    float d=0;
    SD q_nn = {0, FLT_MAX};
	
    do {
        stat_inc_search_loop;
        if (!sbt->seen_for[pj]
			&& sbt->nn[touch_index].max.old + alpha*q_nn.key > d)
			/*       && ((1+2*alpha)*q_nn.key > d  ||   q_nn.point > j-1) */
		{
            d = sbt->dist(q,pj);
            if (d < q_nn.key) {q_nn.key = d; q_nn.point = pj;}
            if (sbt->ins_num[pj] < sbt->end_thresh) {
                MARK_SEEN(sbt, pj);
				sbt->cur_touching[pj] = sbt->offset[pj];
                insert_heap(sbt->psh, pj, -d);
                stat_update_max_heap_entries(sbt->psh->num_entries);
            }
        }
        while (sbt->psh->num_entries>0)
		{
            size_t pm = topp(sbt->psh).point;
            d = -topp(sbt->psh).key;
            touch_index = (sbt->cur_touching[pm])++;
            if (touch_index<sbt->offset[pm+1]
				&& sbt->dDL[touch_index] + alpha*q_nn.key > d)
			{
                pj = sbt->DL[touch_index];
                break;
			}
            delete_heap(sbt->psh, 0);
        }
    } while (sbt->psh->num_entries>0);
	
    reset_aux(sbt);
	
    return q_nn.point;
}





/* here a search_sb_d_order variant */

size_t search_sb_fixed_r(sb* sbt, size_t q, float r, void (*v)(SD*) )
{
	size_t pj=0, touch_index=0, ans=0;
    float d = 0;
    SD qjsd;
		
    do {
        stat_inc_search_loop;
        if (  !sbt->seen_for[pj]
			&& sbt->nn[touch_index].max.old + r > d ) {
            d = sbt->dist(q,pj);
            if (d < r && q!=pj) {ans++; qjsd.point = pj; qjsd.key = d; v(&qjsd);}
			if (sbt->ins_num[pj] < sbt->end_thresh) {
				sbt->cur_touching[pj] = sbt->offset[pj];
				insert_heap(sbt->psh, pj, -d);
				MARK_SEEN(sbt, pj)
				stat_update_max_heap_entries(sbt->psh->num_entries);
			}
		}
        while (sbt->psh->num_entries>0) {
            size_t pm = topp(sbt->psh).point;
            d = -topp(sbt->psh).key;
            touch_index = (sbt->cur_touching[pm])++;
            if (touch_index<sbt->offset[pm+1]
				&& sbt->dDL[touch_index] + r > d) {
                pj = sbt->DL[touch_index];
                break;
            }
            delete_heap(sbt->psh, 0);
        }
    } while (sbt->psh->num_entries>0);
	
    reset_aux(sbt);
	
    return ans;
}


SD nns[1000];
size_t n_nns;

/* like search_sb_i_order, but stop at v, record sequence of
nearest neighbors in nns */
size_t search_sb_up2(sb* sbt, size_t q, size_t v)
{
    size_t j, m, ma;
    float d = 0.0;
	SD q_nn = {0, FLT_MAX};
	heap *psh = sbt->psh;

    j = ma = 0;
	
    while (1) {
        stat_inc_search_loop;
        if (  !sbt->seen_for[j]
			&& sbt->nn[ma].max.old + q_nn.key > d 
			) {
            sbt->dist2q[j] = d = sbt->dist(q,j);
            if (d < q_nn.key) {
                q_nn.key = d; q_nn.point = j;
                nns[n_nns++] = q_nn;
                assert(n_nns<999);
            }
            ma = sbt->cur_touching[j] = sbt->offset[j];
			if (ma<sbt->offset[j+1]) {
				insert_heap(psh, j,-(float)sbt->ins_num[sbt->DL[ma]]);
				MARK_SEEN(sbt, j);
			}
        }
        if (psh->num_entries==0) break;
        do  {
            m = topp(sbt->psh).point;
            d = sbt->dist2q[m];
            ma = (sbt->cur_touching[m])++;
            delete_heap(psh, 0);
            if (ma>=sbt->offset[m+1]
				|| sbt->dDL[ma] + q_nn.key < d
				|| sbt->ins_num[sbt->DL[ma]] > v) continue;
            j = sbt->DL[ma];
            if (ma<sbt->offset[m+1]-1)
                insert_heap(psh,m,-(float)sbt->ins_num[sbt->DL[ma+1]]);
            break;
        } while (psh->num_entries>0);
    }
	
    reset_aux(sbt);
	
    return q_nn.point;
}





static SD *close_sites;
static size_t alloc_close_sites=0, num_close_sites=0;

static void get_close_sites(SD *s) {
	static size_t alloc_close_sites;
	if (num_close_sites>=alloc_close_sites) {
		alloc_close_sites = 1+2*alloc_close_sites;
		close_sites = (SD*)realloc(close_sites, sizeof(SD)*(alloc_close_sites));
	}
	close_sites[num_close_sites++] = *s;
}

#define SD_swap(a, b) {SD t;  t = *(a); *(a) = *(b);  *(b) = t;}                    


/* assuming 0<k<=m */

static void select_k_nearest(SD *cs, size_t k, size_t m) {
	int i, j;
	
	if (k>=m)
		return;

	for (i=1, j=m-1; i<j;) {
		while (i<j && cs[i].key <= cs[0].key) i++;
		while (i<j && cs[j].key > cs[0].key) j--;
		SD_swap(cs+i, cs+j);
	}
	
	if (cs[i].key <= cs[0].key)
		i++; /* now cs[h].key <= cs[0].key iff h<i */

	SD_swap(cs+i-1, cs);

	if (i==(int)k)
		return;

	if (i>(int)k)
		select_k_nearest(cs, k, i-1); /* don't need pivot */
	else
		select_k_nearest(cs+i, k-i, m-i); /* i>0 implies progress */
}


void search_sb_k_nearest(sb* sbt, size_t q, size_t k, void (*v)(SD*))
{
    size_t j, i = sbt->n/(k+1);
    float r;
    SD s;
		
    n_nns = 0;
    search_sb_up2(sbt, q, i);
	
    while (1) {
        if (n_nns==0) {
            num_close_sites = 0;
            for (i=0; i<sbt->n; i++) {
                s.key = sbt->dist(i, q); s.point = i;
				if (s.key>0)
					get_close_sites(&s);
            }
            r = FLT_MAX;
            break;
        }
		num_close_sites = 0;
        j = search_sb_fixed_r(sbt, q, r = nns[--n_nns].key, &get_close_sites);
        if (0) printf("r=%f, j=%d, n_nns=%d\n", r, j, n_nns);
        if (j>k) break;
    }
    select_k_nearest(close_sites, k, num_close_sites);
	if (k>num_close_sites)
		k = num_close_sites;
	for (j=0; j<k; j++)
		v(&close_sites[j]);
}
